#include "numericalpack.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

void _MW_swapIndex(int index1, int index2, double *A) {
  double tmp = A[index1];
  A[index1] = A[index2];
  A[index2] = tmp;
}


/**
 * Swap Rows of A. rowIndex1 and rowIndex2 gives the index to the 
 * an element in each row. The matrix A is of size m x n.
 */
void _MW_swapRows(int rowIndex1, int rowIndex2, int m, int n, double *A) 
{
  int i, row1, row2; double tmp[n];

  /* figure out which rows to interchange */
  row1 = n*((int) floor( (double) rowIndex1/ (double) m));
  row2 = n*((int) floor( (double) rowIndex2/ (double) m));

  for (i = 0; i < n; i ++) {
    tmp[i] = A[row1 + i];
    A[row1 + i] = A[row2 + i];
    A[row2 + i] = tmp[i];
  }
}

int MW_LUfactorization(int m, int n, double *A, double *L)
{
  int i,j,k;

  /* initialize L */
  for (i = 0; i < m*n; i++)
    L[i] = 0.0;

  for (i = 0; i < n; i++) {
    L[i*m + i] = (double) 1.0;
  }
  

  for (i = 0; i < n; i++) {
    for (j = i+1; j < m; j++) {
      if (A[i*m+i] == 0)
	return 0;  /* singular matrix */

      L[j*m + i] = A[j*m + i] / A[i*m + i];
      A[j*m+i] = 0;     


      for (k = i+1; k < m; k++) {
	A[j*m + k] = A[j*m + k] - L[j*m+i] * A[i*m + k];
      }
    }
  } 


  return 1;
}


int MW_PLUfactorization(int m, int n, double *A, double *L, double *P)
{
  int i,j,k, max_index;
  double max = 0.0; 
  char *str = "ERROR MW_PLUfactorization:";


  /* initialize L */
  for (i = 0; i < m*n; i++)
    L[i] = 0.0;
  
  /* initialize P */
  for (i = 0; i < m*n; i++) {
    P[i] = 0.0;
  }

  for (i = 0; i < n; i++) {
    P[i*m + i] = (double) 1.0;
  }
  
  /* begining main routine */
  for (k = 0; k < n-1; k++) {

    /* search after largest element in column k */
    for (i = k; i < m; i ++) {
      if (fabs(A[i*n + k]) > max) {
	max = fabs(A[i*n+k]);
	max_index = i*n+k;
      }
    }

    if (max == 0) {
      printf("%s A is singular\n",str);
      return 0;
    }

    if (max_index > k*n + k) {
      /* make a swap in the pivot matrix */
      _MW_swapRows(k*n+k, max_index, m, n, P);

      /* swap rows of A */
      _MW_swapRows(k*n+k, max_index, m, n, A); 

      /* swap rows of L */
      _MW_swapRows(k*n+k, max_index, m, n, L); 
    }

    /* Gaussian elimination */
    for (i = k+1; i < m; i++) {
      L[i*n + k] = A[i*n + k] / A[k*n + k];
      A[i*n + k] = 0.0;
      for (j = 0; j < m-(k+1); j++) {
	A[i*n+k+1+j] = A[i*n+k+1+j] - L[i*n+k] * A[k*n+k+1+j];
      }
    } 

    
    /* put ones on the main diagonal of L */
    for (i = 0; i < n; i++) {
      L[i*m + i] = (double) 1.0;
    }


    max = 0.0; /* reset max */
  }

  return 1;
}



void MW_displayMatrix(int m, int n, double *A)
{
  int i, j;
  
  printf("\n");
  for (i=0; i < m; i++) {
    for (j=0; j < n; j++) {
      printf("%8.5f ", (float) A[i*m +j]);
    }
    printf("\n");
    fflush(NULL);
  }
}


/**
 * Converts matrix from row major mode to column major mode
 * as used in e.g. OpenGL.
 */

void MW_transpose(int m, int n, double *A, double *B)
{
  int i, j;
  
  for (i = 0; i < m; i ++) {
    for (j = 0; j < n; j++) {
      B[j*n + i] = A[i*m + j];
    }
  }
}


int MW_inverse(int n, double *P, double *L, double *U, double *Ainv)
{
  int i,j,k;
  double back_l;

  /* Initialize Ainv */
  for (i=0; i < n*n; i++)
    Ainv[i] = P[i];

  /* Update Ainv, by using elimination coefficients */
  for (k=0; k < n-1; k++) {      
    /* Gaussian elimination */
    for (i = k+1; i < n; i++) {
      for (j = 0; j < n; j++) {
	Ainv[i*n+j] = Ainv[i*n+j] - L[i*n+k] * Ainv[k*n+j];
      }
    } 
  }

  /* Update Ainv so that U is in row-echelon form
     a kind of backsubstitution where U is only a 
     diagonal.
  */
  for(k=n-1; k > 0; k--) {
    for (i=k-1; i > -1; i--) {
      if (U[i*n+k] != 0) {
	back_l = U[i*n+k]/U[k*n+k];
	//printf("L = A[%d] / A[%d]\n", i*n+k, k*n+k);
	for (j = n-1; j > -1; j--) {
	  Ainv[i*n+j] = Ainv[i*n+j] - back_l*Ainv[k*n+j];
	  //printf("Ainv[%d] - L * Ainv[%d]\n", i*n+j, k*n+j);
	}
      }
    }
  }

  /* Scale each column in Ainv with diagonal element of U */
  for (k=0; k < n; k++) {
    for (i=0; i < n; i++) {
      Ainv[k*n + i] = Ainv[k*n + i] / U[k*n + k];
    }
  }

  return 1;
}


